/*
 * AbstractHeap.java:  Nov 3, 2006
 * Changxi Zheng (cxzheng@cs.cornell.edu)
 */
package cornell.cs211;

import java.util.Collection;
import java.util.Iterator;
import java.util.Comparator;
import java.util.AbstractCollection;

/**
 * This class provides skeletal implementations of some {@link Heap}
 * operations. It is integrated with 
 * <a href="http://java.sun.com/j2se/1.5.0/docs/guide/collections/index.html">
 * Java Collection framework</a>
 * 
 * <p> A <tt>Heap</tt> implementation that extends this class must 
 * miniumally define methods <tt>push</tt>, <tt>pop</tt> and <tt>top</tt>
 * 
 * @see Heap
 * @author Changxi Zheng 
 * @param <E> The type of the elements held in this collection.
 */
public abstract class AbstractHeap<E> 
        extends AbstractCollection<E> 
        implements Heap<E> {

    /**
     * The comparator, or null if priority queue uses elements'
     * natural ordering.
     */
    protected final Comparator<? super E> comparator;

    /**
     * Construct with the given comparator which is used to decide the 
     * order of elements in the heap
     */
    protected AbstractHeap(Comparator<? super E> comparator) {                
        if ( comparator != null ) 
            this.comparator = comparator;
        else
            this.comparator = new Cmp();
    }
    /**
     * Adds the specified element to this heap.
     * 
     * @return <tt>true</tt> if add successfully, <tt>false</tt> otherwise.
     */
    public boolean add(E o) {
       return push(o);   
    }
    
    /**
     * Remove all the elements from the heap.
     */
    public void clear() {
        while ( pop()!= null );
    }
    
    /**
     * Adds all of the elements in the specified collection to this queue.
     * The current implementation is just to add the elements in the given <tt>
     * Collection</tt> instance one by one.
     *
     * @param c collection whose elements are to be added into this heap.
     * @return <tt>true</tt> if this collection changed as a result of this call.
     * @throws NullPointerException if the specified collection is null.
     * @throws IllegalArgumentException if the specified collection is this heap.
     */
    public boolean addAll(Collection<? extends E> c) {
        if ( c == null ) 
            throw new NullPointerException();
        if ( c == this ) 
            throw new IllegalArgumentException();
        boolean modified = false;
        Iterator<? extends E> e = c.iterator();
        while ( e.hasNext() ) {
            if ( push(e.next()) ) 
                modified = true;
        }
        return modified;
    }
    
    /**
     * Default comparator
     */
    private class Cmp implements Comparator<E> {
        public int compare(E o1, E o2) {
            return ((Comparable<E>)o1).compareTo(o2);
        }

        public boolean equals(Object obj) {
            return false;
        }
    }
}
