/*
 * SortedListHeap.java:  Nov 3, 2006
 * Changxi Zheng (cxzheng@cs.cornell.edu)
 */
package cornell.cs211;

import java.util.ArrayList;
import java.util.Comparator;
import java.util.LinkedList;
import java.util.List;
import java.util.Iterator;

/**
 * The implementation of the heap using the sorted list. The internal 
 * storage can be any list instances, i.e. linked list or array list
 *
 * <p>Implementation note: this implementation provides O(1) time
 * for the methods <tt>top</tt>, <tt>pop</tt>, but O(n) for
 * <tt>push</tt>, <tt>add</tt> methods.
 * 
 * @author Changxi Zheng
 *
 * @param <E> The type of the elements held in this heap.
 */
public class SortedHeap<E> extends AbstractHeap<E> {
    /**
     * Stores the elements in the heap
     */
    private List<E> list = null;
    
    /**
     * Creates a <tt>SortedHeap</tt> that orders its elements according to 
     * their natural ordering (using <tt>Comparable</tt>).
     */
    public SortedHeap(List<E> list) {
        this(list, null);        
    }

    /**
     * Creates a <tt>SortedHeap</tt> that orders its elements according to 
     * the specified comparator.
     * 
     * @param comparator the comparator used to order this heap.
     *        If <tt>null</tt> then the order depends on the elements' natural
     *        ordering. 
     */
    public SortedHeap(List<E> list, Comparator<? super E> comparator) {        
        super(comparator);        
        this.list = list;        
    }    
    
    /**
     * Factory Method to create a <tt>SortedHeap</tt> using <tt>ArrayList</tt>
     */
    static public <T extends Object> SortedHeap<T> sortedArrayHeap(T d) {
        return new SortedHeap<T>(new ArrayList<T>());
    }
    
    /**
     * Factory Method to create a <tt>SortedHeap</tt> using <tt>LinkedList</tt>
     */
    static public <T extends Object> SortedHeap<T> sortedListHeap(T d) {
        return new SortedHeap<T>(new LinkedList<T>());
    }
    
    /**
     * Factory Method to create a <tt>SortedHeap</tt> using <tt>LinkedList</tt> with the
     * specified comparator to determine the order in the heap
     */
    static public <T extends Object> SortedHeap<T> 
            sortedListHeap(Comparator<? super T> comparator) {
        return new SortedHeap<T>(new LinkedList<T>(), comparator);
    }
    
    /**
     * Factory Method to create a <tt>SortedHeap</tt> using <tt>ArrayList</tt> with the
     * specified comparator to determine the order in the heap
     */
    static public <T extends Object> SortedHeap<T> 
            sortedArrayHeap(Comparator<? super T> comparator) {
        return new SortedHeap<T>(new ArrayList<T>(), comparator);
    }
    
    public boolean push(E o) {
        if ( o == null ) 
            throw new NullPointerException();
        
        Iterator<E> iter = list.iterator();
        int pos = 0;
        while ( iter.hasNext() ) {
            if ( comparator.compare(o, iter.next()) <= 0 ) {
                list.add(pos, o);
                return true;
            }
            ++ pos;
        }        
        list.add(list.size(), o);
        return true;        
    }
    
    public E top() {
        if ( list.isEmpty() )
            return null;        
        return list.get(0);
    }
    
    public E pop() {
        if ( list.isEmpty() )
            return null;        
        return list.remove(0);        
    }
    
    /**
     * Removes all elements from the heap.
     * The heap will be empty after this call returns.
     */
    public void clear() {
        list.clear();
    }
    
    /**
     * Returns the current size.
     */
    public int size() {
        return list.size();
    }
    
    /**
     * Returns an iterator over the elements in this heap. 
     * 
     * @return an iterator over the elements in this heap.
     */
    public Iterator<E> iterator() {
        return list.iterator();
    }
}
