/*
 * StdHeap.java:  Nov 3, 2006
 * Changxi Zheng (cxzheng@cs.cornell.edu)
 */
package cornell.cs211;

import java.util.Iterator;
import java.util.Comparator;
import java.util.NoSuchElementException;

/**
 * The standard implementation of the heap, which is very similar
 * with <tt>java.util.PriorityQueue</tt>, using the binary serach tree (BST).
 * 
 * <p>The BST used in this class is 
 *
 * <p>Implementation note: this implementation provides O(log(n)) time
 * for the insertion methods (<tt>push</tt>, <tt>pop</tt>,
 * <tt>remove()</tt> and <tt>add</tt>) methods; linear time for the
 * <tt>remove(Object)</tt> and <tt>contains(Object)</tt> methods; and
 * constant time for the retrieval methods (<tt>peek</tt>,
 * <tt>element</tt>, and <tt>size</tt>).
 * 
 * @author Changxi Zheng
 * @param <E> The type of the elements held in this collection.
 */
public class StdHeap<E> extends AbstractHeap<E> {

    private static final int DEFAULT_INIT_CAPACITY = 11;
    
    /**
     * Number of elements in the current heap.
     */
    private int size = 0;

    /**
     * Priority heap represented as a balanced binary heap: the two children
     * of heap[n] are heap[2*n] and heap[2*n + 1].
     */
    private transient Object[] heap;

    /**
     * Creates a <tt>StdHeap</tt> with the default initial capacity
     * (11) that orders its elements according to their natural
     * ordering (using <tt>Comparable</tt>).
     */
    public StdHeap() {
        this(DEFAULT_INIT_CAPACITY, null);
    }

    /**
     * Creates a <tt>StdHeap</tt> with the specified initial capacity
     * that orders its elements according to the specified comparator.
     *
     * @param initCapacity the initial capacity for this heap.
     * @param comparator the comparator used to order this heap.
     * If <tt>null</tt> then the order depends on the elements' natural
     * ordering.
     * @throws IllegalArgumentException if <tt>initialCapacity</tt> is less
     * than 1
     */
    public StdHeap(int initCapacity, Comparator<? super E> comparator) {        
        super(comparator);
        if ( initCapacity < 1 ) 
            throw new IllegalArgumentException();
        this.heap = new Object[initCapacity + 1];        
    }

    /**
     * Inserts the specified element into this heap.
     *
     * @return <tt>true</tt>
     * @throws ClassCastException if the specified element cannot be compared
     * with elements currently in the heap according to the heap's ordering.
     * @throws NullPointerException if the specified element is <tt>null</tt>.
     */ 
    public boolean push(E o) {
        if ( o == null ) 
            throw new NullPointerException();
        ++ size;

        if ( size >= heap.length ) growSize(size);
        heap[size] = o;
        fixUp(size);
        return true;
    }
    
    /**
     * Removes all elements from the heap.
     * The heap will be empty after this call returns.
     */
    public void clear() {
        // Null out element references to prevent memory leak
        for (int i=1; i<=size; heap[i++]=null);
        size = 0;
    }

    /**
     * Returns the current size.
     */
    public int size() {
        return size;
    }

    public E top() {
        if ( size==0 )
            return null;
        return (E)heap[1];
    }
    
    public E pop() {
        E ret = (E)heap[1];
        heap[1] = heap[size];
        heap[size--] = null;
        if ( size > 1 )
            fixDown(1);
        return ret;
    }
    /**
     * Returns an iterator over the elements in this heap. The iterator
     * does not return the elements in any particular order.
     * 
     * <p>
     * Currently, The iterator doesn't support <tt>remove</tt> method, 
     * since it is a optional in <tt>Iterable</tt> interface.
     *
     * @return an iterator over the elements in this heap.
     */
    public Iterator<E> iterator() {
        return new Itr();
    }
    
    /*=================== Private Methods =======================*/
    private void fixUp(int k) {
        while (k > 1) {
            int j = k >> 1;
            if (comparator.compare((E)heap[j], (E)heap[k]) <= 0)
                break;
            Object tmp = heap[j];  heap[j] = heap[k]; heap[k] = tmp;
            k = j;
        }        
    }
    
    private void fixDown(int k) {
        int j;
        while ((j = k << 1) <= size && (j > 0)) {
            if (j<size && 
                comparator.compare((E)heap[j], (E)heap[j+1]) > 0)
                j++; // j indexes smallest kid
            if (comparator.compare((E)heap[k], (E)heap[j]) <= 0)
                break;
            Object tmp = heap[j];  heap[j] = heap[k]; heap[k] = tmp;
            k = j;
        }        
    }
    
    private E removeAt(int i) {
        E ret = (E)heap[i];
        heap[i] = heap[size];
        heap[size--] = null;  // Drop extra ref to prevent memory leak
        if (i <= size) {
            fixDown(i);   
            fixUp(i);
        }
        return ret;
    }
        
    private void growSize(int index) {
        int newLen = heap.length;
        if ( index < newLen )
            return;
        if ( index == Integer.MAX_VALUE)
            throw new OutOfMemoryError();
        while ( newLen <= index ) {
            if ( newLen >= Integer.MAX_VALUE / 2 ) // avoid overflow
                newLen = Integer.MAX_VALUE;
            else 
                newLen <<= 2;
        }
        Object[] newHeap = new Object[newLen];
        System.arraycopy(heap, 0, newHeap, 0, heap.length);
        heap = newHeap;
    }
    
    private class Itr implements Iterator<E> {
        
        private int cursor = 1;
        
        public boolean hasNext() {
            return cursor <= size;
        }

        public E next() {            
            if (cursor <= size) {
                return (E)heap[cursor++];
            } else 
                throw new NoSuchElementException();
        }

        public void remove() {
            throw new UnsupportedOperationException();
        }
    }
}
