/*
 * UnsortedHeap.java:  Nov 3, 2006
 * Changxi Zheng (cxzheng@cs.cornell.edu)
 */
package cornell.cs211;

import java.util.Comparator;
import java.util.Iterator;
import java.util.List;
import java.util.LinkedList;
import java.util.ArrayList;

/**
 * The implementation of the heap using the unsorted list. The internal 
 * storage can be any list instances, i.e. linked list or array list
 *
 * <p>Implementation note: this implementation provides O(1) time
 * for the insertion methods (<tt>push</tt>, <tt>add</tt>), but O(n) for
 * <tt>remove()</tt>, <tt>top</tt> and <tt>pop</tt> methods.
 * 
 * @author Changxi Zheng
 * @param <E> The type of the elements held in this collection.
 */
public class UnsortedHeap<E> extends AbstractHeap<E> {
    
    /**
     * Stores the elements in the heap
     */
    private List<E> list = null;
    
    /**
     * Creates a <tt>UnsortedHeap</tt> that orders its elements according to 
     * their natural ordering (using <tt>Comparable</tt>).
     */
    public UnsortedHeap(List<E> list) {
        this(list, null);        
    }

    /**
     * Creates a <tt>UnsortedHeap</tt> that orders its elements according to 
     * the specified comparator.
     * 
     * @param comparator the comparator used to order this heap.
     *        If <tt>null</tt> then the order depends on the elements' natural
     *        ordering. 
     */
    public UnsortedHeap(List<E> list, Comparator<? super E> comparator) {        
        super(comparator);        
        this.list = list;        
    }    
        
    /**
     * Factory Method to create a <tt>UnsortedHeap</tt> using <tt>ArrayList</tt>
     */
    static public <T extends Object> UnsortedHeap<T> unsortedArrayHeap(T d) {
        return new UnsortedHeap<T>(new ArrayList<T>());
    }
    
    /**
     * Factory Method to create a <tt>UnsortedHeap</tt> using <tt>LinkedList</tt>
     */
    static public <T extends Object> UnsortedHeap<T> unsortedListHeap(T d) {
        return new UnsortedHeap<T>(new LinkedList<T>());
    }
    
    /**
     * Factory Method to create a <tt>UnsortedHeap</tt> using <tt>LinkedList</tt> with the
     * specified comparator to determine the order in the heap
     */
    static public <T extends Object> UnsortedHeap<T> 
            unsortedListHeap(Comparator<? super T> comparator) {
        return new UnsortedHeap<T>(new LinkedList<T>(), comparator);
    }
    
    /**
     * Factory Method to create a <tt>UnsortedHeap</tt> using <tt>ArrayList</tt> with the
     * specified comparator to determine the order in the heap
     */
    static public <T extends Object> UnsortedHeap<T> 
            unsortedArrayHeap(Comparator<? super T> comparator) {
        return new UnsortedHeap<T>(new ArrayList<T>(), comparator);
    }
    /**
     * Inserts the specified element into this heap.
     * As this is a unsorted heap, the item is attached to the end of list. So
     * the time complexity is O(1)
     *
     * @return <tt>true</tt>
     * @throws ClassCastException if the specified element cannot be compared
     * with elements currently in the heap according to the heap's ordering.
     * @throws NullPointerException if the specified element is <tt>null</tt>.
     */ 
    public boolean push(E o) {
        if ( o == null ) 
            throw new NullPointerException();        
        return list.add(o);        
    }
    
    /**
     * Removes all elements from the heap.
     * The heap will be empty after this call returns.
     */
    public void clear() {
        list.clear();
    }
    
    /**
     * Returns the current size.
     */
    public int size() {
        return list.size();
    }

    public E top() {
        if ( list.isEmpty() )
            return null;
        Iterator<E> iter = list.iterator();
        E ret = iter.next();
        while ( iter.hasNext() ) {
            E c = iter.next();
            if ( comparator.compare(ret, c) > 0 )
                ret = c;
        }
        return ret;
    }
    
    public E pop() {
        if ( list.isEmpty() )
            return null;
        Iterator<E> iter = list.iterator();
        E ret = iter.next();
        while ( iter.hasNext() ) {
            E c = iter.next();
            if ( comparator.compare(ret, c) > 0 )
                ret = c;
        }
        list.remove(ret);
        return ret;
    }
    
    /**
     * Returns an iterator over the elements in this heap. The iterator
     * does not return the elements in any particular order.
     * 
     * @return an iterator over the elements in this heap.
     */
    public Iterator<E> iterator() {
        return list.iterator();
    }
}
